import java.util.*;
import java.io.*;

public class Slice
{
   // Constants;
   public static final boolean DEBUG = false;
   public static final String outputFileName = "SLICE.txt";
   
   // Flags for options specified on the command line.
   public static boolean limitedTraceFile = false;
   
   // Data structures.
   public static Hashtable<String,InstanceInfo> instanceInfoTable = new Hashtable<String,InstanceInfo>();
   
   public static void main(String[] args) throws Exception
   {
      // Process input arguments.
      if(args.length < 3)
      {
         fatalError("Usage: java Slice [options] TRACE_FILE_NAME INSTRUCTION_TO_SLICE_FROM INSTANCE_TO_SLICE_FROM\n" + 
                    "   options:\n" +
                    "      --limited:  slice on a trace file that is specified using\n" +
                    "                  the \"limited execution history\" format");
      }
      
      debug("Processing arguments...");
      
      for(int i = 0; i < args.length-3; i++)
      {
         processOption(args[i]);
      }
      String inputFileName = args[args.length-3];
      String instrToSliceFrom = args[args.length-2];
      String instToSliceFrom = args[args.length-1];
      
      File tempFile = new File(inputFileName);
      myAssert(tempFile.exists(), "Specified trace file \"" + inputFileName + "\" does not exist");
      try
      {
         if(instrToSliceFrom.indexOf("0x") == 0)
         {
            instrToSliceFrom = instrToSliceFrom.substring(2);
         }
         Long.parseLong(instrToSliceFrom, 16);
      }
      catch(Exception e)
      {
         fatalError("Invalid instruction address specified: \"" + instrToSliceFrom + "\" (must be a hex number)");
      }
      try
      {
         Integer.parseInt(instToSliceFrom);
      }
      catch(Exception e)
      {
         fatalError("Invalid instance number specified: \"" + instToSliceFrom + "\" (must be an integer number)");
      }
      
      if(instrToSliceFrom.indexOf("0x") != 0)
      {
         instrToSliceFrom = "0x" + instrToSliceFrom;
      }
      
      printOptions();
      debug("Input file name is: " + inputFileName);
      debug("Instruction to slice from: " + instrToSliceFrom);
      debug("Instance number to slice from: " + instToSliceFrom);
      
      // Read trace file to collect dependence information.
      if(limitedTraceFile)
      {
         readDependencesLimited(inputFileName);
      }
      else
      {
         readDependencesComprehensive(inputFileName);
      }
      
      // Perform the slicing.
      Hashtable<String,Boolean> slice = sliceFrom(instrToSliceFrom, instToSliceFrom);
      
      // Output the results.
      outputSlicingResults(slice);
   }
   
   public static void outputSlicingResults(Hashtable<String,Boolean> slice) throws Exception
   {
      Vector<SliceEntry> sliceSorted = new Vector<SliceEntry>();
      Enumeration<String> en = slice.keys();
      while(en.hasMoreElements())
      {
         String entry = en.nextElement();
         String part1 = entry.substring(0, entry.indexOf("_"));
         String part2 = entry.substring(entry.indexOf("_")+1);
         SliceEntry sliceEntry = new SliceEntry(part1, part2);
         sliceSorted.add(sliceEntry);
      }
      Collections.sort(sliceSorted);
      
      FileWriter out = new FileWriter(outputFileName);
      out.write("Computed slice: " + sliceSorted.size() + "/" + instanceInfoTable.size() + " instruction instances in slice/trace\n");
      out.write("Instruction instances in slice:\n");
      for(int i = 0; i < sliceSorted.size(); i++)
      {
         SliceEntry entry = sliceSorted.get(i);
         out.write("   " + entry.instr + "#" + entry.inst + "\n");
      }
      out.close();
      
      System.out.println("Computed slice: " + sliceSorted.size() + "/" + instanceInfoTable.size() + " instruction instances in slice/trace");
      System.out.println("Slice written to output file: \"" + outputFileName + "\"");
   }
   
   public static Hashtable<String,Boolean> sliceFrom(String instrStart, String instStart)
   {
      InstanceInfo ii = instanceInfoTable.get(instrStart+"_"+instStart);
      myAssert(ii != null, "Slicing error: specified starting instruction " + instrStart + ", instance " + instStart + ", is not in the trace file");
      Hashtable<String,Boolean> result = new Hashtable<String,Boolean>();
      return sliceRecursive(result, ii);
   }
   
   public static Hashtable<String,Boolean> sliceRecursive(Hashtable<String,Boolean> result, InstanceInfo ii)
   {
      if(result.get(ii.instr+"_"+ii.inst) != null)
      {
         // This instruction instance is already in the result, so it has already been processed before.
         return result;
      }
      
      result.put(ii.instr+"_"+ii.inst, new Boolean(true));
      
      Hashtable<String,Boolean> dependsUpon = ii.dependsUpon;
      Enumeration<String> en = dependsUpon.keys();
      while(en.hasMoreElements())
      {
         String dep = en.nextElement();
         InstanceInfo ii2 = instanceInfoTable.get(dep);
         if(ii2 != null)
         {
            sliceRecursive(result, ii2);
         }
      }
      
      return result;
   }
   
   public static void readDependencesLimited(String inputFileName) throws Exception
   {
      BufferedReader infile = new BufferedReader(new FileReader(inputFileName));
      int lineNum = 0;
      while(infile.ready())
      {
         String line = infile.readLine().trim();
         lineNum++;
         if(line.equals("")) continue;
         myAssert(line.indexOf("0x") >= 0, "Unrecognized format at line " + lineNum + " in trace input file:\n   \"" + line + "\"");
         myAssert(line.indexOf(" --> ") >= 0, "Unrecognized format at line " + lineNum + " in trace input file:\n   \"" + line + "\"");
         myAssert(line.indexOf("#") >= 0, "Unrecognized format at line " + lineNum + " in trace input file:\n   \"" + line + "\"");
         
         String part1 = line.substring(0, line.indexOf(" "));
         String part2 = line.substring(line.indexOf("> ")+2);
         
         String instr1 = part1.substring(0, part1.indexOf("#"));
         String inst1 = part1.substring(part1.indexOf("#")+1);
         String instr2 = part2.substring(0, part2.indexOf("#"));
         String inst2 = part2.substring(part2.indexOf("#")+1);
         
         InstanceInfo ii = instanceInfoTable.get(instr1 + "_" + inst1);
         if(ii == null)
         {
            ii = new InstanceInfo(instr1, inst1);
            instanceInfoTable.put(instr1 + "_" + inst1, ii);
         }
         ii.addDependsUpon(instr2, inst2);
         
         ii = instanceInfoTable.get(instr2 + "_" + inst2);
         if(ii == null)
         {
            ii = new InstanceInfo(instr2, inst2);
            instanceInfoTable.put(instr2 + "_" + inst2, ii);
         }
      }
      infile.close();
   }
   
   public static void readDependencesComprehensive(String inputFileName) throws Exception
   {
      BufferedReader infile = new BufferedReader(new FileReader(inputFileName));
      int lineNum = 0;
      String numBlocksString = infile.readLine().trim();
      try
      {
         Integer.parseInt(numBlocksString);
      }
      catch(Exception e)
      {
         fatalError("First line of comprehensive trace file must be a single integer");
      }
      int numBlocks = Integer.parseInt(numBlocksString);
      Hashtable<String,String> idToInstructionAddress = new Hashtable<String,String>();
      for(int i = 0; i < numBlocks; i++)
      {
         String staticInfo = infile.readLine().trim();
         StringTokenizer tok = new StringTokenizer(staticInfo);
         String id = tok.nextToken();
         String numUsePortsString = tok.nextToken();
         String instructionAddress = "0x" + tok.nextToken();
         
         int numUsePorts = 0;
         try
         {
            numUsePorts = Integer.parseInt(numUsePortsString);
         }
         catch(Exception e)
         {
            fatalError("Invalid value for number of use ports in input line: \"" + staticInfo + "\"");
         }
         
         idToInstructionAddress.put(id, instructionAddress);
         
         for(int j = 0; j < numUsePorts; j++)
         {
            String sizeLine = infile.readLine().trim();
            myAssert(sizeLine.indexOf("SIZE") == 0, "Unrecognized SIZE line in specified trace file: \"" + sizeLine + "\"");
            String sizeVal = sizeLine.substring(sizeLine.indexOf(" ")+1);
            try
            {
               Integer.parseInt(sizeVal);
            }
            catch(Exception e)
            {
               fatalError("Unrecognized SIZE value in specified trace file: \"" + sizeLine + "\"");
            }
            int size = Integer.parseInt(sizeVal);
            for(int k = 0; k < size; k++)
            {
               String depLine = infile.readLine().trim();
               String currInstance = depLine.substring(0, depLine.indexOf(":"));
               myAssert(currInstance.equals(""+k), "Unrecognized format in dependence line from trace file: \"" + depLine + "\"");
               String restOfLine = depLine.substring(depLine.indexOf(":")+1);
               String depID = restOfLine.substring(0, restOfLine.indexOf(" "));
               String depInstanceNum = restOfLine.substring(restOfLine.indexOf(" ")+1);
               
               // Add the dependence info (afterwards, the instruction ID needs to be converted to
               // the actual instruction address, once we know what they all are).
               InstanceInfo ii = instanceInfoTable.get(instructionAddress + "_" + currInstance);
               if(ii == null)
               {
                  ii = new InstanceInfo(instructionAddress, currInstance);
                  instanceInfoTable.put(instructionAddress + "_" + currInstance, ii);
               }
               ii.addDependsUpon(depID, depInstanceNum);
            }
         }
         
         String nextLine = infile.readLine().trim();
         if(nextLine.indexOf("VALUES") == 0)
         {
            String valueCountString = nextLine.substring(nextLine.indexOf(" ")+1);
            int valueCount = 0;
            try
            {
               valueCount = Integer.parseInt(valueCountString);
            }
            catch(Exception e)
            {
               fatalError("Invalid value for number of value entries in input line: \"" + nextLine + "\"");
            }
            for(int j = 0; j < valueCount; j++)
            {
               infile.readLine(); // Just bypass the values line.
            }
         }
         else if(nextLine.indexOf("NO VALUES") == 0)
         {
            // Just go onto the next instruction block.
         }
         else
         {
            fatalError("Unexpected VALUES line in specified trace file: \"" + nextLine + "\"");
         }
      }
      infile.close();
      
      // Now go through all the collected dependence info and convert each instruction block ID into the
      // corresponding actual instruction address.
      Enumeration<String> en = instanceInfoTable.keys();
      while(en.hasMoreElements())
      {
         InstanceInfo ii = instanceInfoTable.get(en.nextElement());
         Hashtable<String,Boolean> depUpon = ii.dependsUpon;
         Hashtable<String,Boolean> newDepUpon = new Hashtable<String,Boolean>();
         
         Enumeration<String> en2 = depUpon.keys();
         while(en2.hasMoreElements())
         {
            String depEntry = en2.nextElement();
            String tempID = depEntry.substring(0, depEntry.indexOf("_"));
            if(tempID.equals("-1")) continue;
            String tempInstance = depEntry.substring(depEntry.indexOf("_")+1);
            String tempAddress = idToInstructionAddress.get(tempID);
            myAssert(tempAddress != null, "Internal error: instruction block ID " + tempID + " has no associated instruction address");
            newDepUpon.put(tempAddress+"_"+tempInstance, new Boolean(true));
            
            InstanceInfo ii2 = instanceInfoTable.get(tempAddress + "_" + tempInstance);
            if(ii2 == null)
            {
               ii2 = new InstanceInfo(tempAddress, tempInstance);
               instanceInfoTable.put(tempAddress + "_" + tempInstance, ii2);
            }
         }
         
         ii.dependsUpon = newDepUpon;
      }
   }
   
   public static void processOption(String option)
   {
      if(option.equals("--limited"))
      {
         limitedTraceFile = true;
      }
      else
      {
         fatalError("Unrecognized option: " + option);
      }
   }
   
   public static void printOptions()
   {
      debug("Trace file in limited execution history format? " + limitedTraceFile);
   }
   
   public static void myAssert(boolean ok, String message)
   {
      if(!ok)
      {
         fatalError(message);
      }
   }
   
   public static void debug(String message)
   {
      if(DEBUG)
      {
         System.out.println(message);
      }
   }
   
   public static void fatalError(String message)
   {
      System.out.println("=== Slicing tool terminating with a fatal error ===");
      System.out.println(message);
      System.out.println("=== Slicing tool terminated ===");
      System.exit(0);
   }
   
   public static class SliceEntry implements Comparable<SliceEntry>
   {
      String instr;
      String inst;
      
      public SliceEntry(String instr, String inst)
      {
         this.instr = instr;
         this.inst = inst;
      }
      
      public int compareTo(SliceEntry other)
      {
         myAssert(other.instr.indexOf("0x") == 0, "Internal error: wrong format expected in string 1");
         myAssert(this.instr.indexOf("0x") == 0, "Internal error: wrong format expected in string 2");
         String tempOther = other.instr.substring(2);
         String tempThis = this.instr.substring(2);
         long otherLong = Long.parseLong(tempOther, 16);
         long thisLong = Long.parseLong(tempThis, 16);
         if(thisLong < otherLong)
         {
            return -1;
         }
         else if(thisLong > otherLong)
         {
            return 1;
         }
         else
         {
            int otherInt = Integer.parseInt(other.inst);
            int thisInt = Integer.parseInt(this.inst);
            if(thisInt < otherInt)
            {
               return -1;
            }
            else if(thisInt > otherInt)
            {
               return 1;
            }
            else
            {
               return 0;
            }
         }
      }
   }
   
   public static class InstanceInfo
   {
      String instr;
      String inst;
      
      Hashtable<String,Boolean> dependsUpon;
      
      public InstanceInfo(String instr, String inst)
      {
         this.instr = instr;
         this.inst = inst;
         this.dependsUpon = new Hashtable<String,Boolean>();
      }
      
      public void addDependsUpon(String instr, String inst)
      {
         this.dependsUpon.put(instr+"_"+inst, new Boolean(true));
      }
   }
}
